/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#ifndef __HASH_H__
#define __HASH_H__

#define NETL_MAX_HASH_BITS 15

#include <assert.h>
struct netl_hash_entry {
	uint32_t hash; /* hash value for the above key */
	struct netl_hash_entry	*next;
};

struct netl_hash_table {
	struct netl_hash_entry **hash_table;
	spinlock_t hash_table_lock;
	int size;
	uint32_t mask;
	int (*compare)(struct netl_hash_entry *cur, struct netl_hash_entry *entry);
	uint32_t (*hash)(const char *str, int len);
};

/*
 * Use simple string hash, a hash_node structure defined to contain
 * pointer to data block, hash, and a ptr to the next node in the chain.
 *
 * Hash table and the free hash nodes are statically allocated.
 *
 * Keep the api simple so that it can be replaced with any other collection
 * later - using trees is not worth the complexity now.
 *
 * Hash insert will return a unique id , we use (hash_array_index) << 15 ored
 * with an empty index in the table. So with the unique id we can identify
 * the hash array index
 */

static inline int netl_hash_table_init(struct netl_hash_table *t, int hash_bits, 
		int (*compare)(struct netl_hash_entry *cur, struct netl_hash_entry *entry))
{
	int size;
	assert(t);
	assert(hash_bits < NETL_MAX_HASH_BITS);

	size = 1<<hash_bits;
	spin_lock_init(&t->hash_table_lock);
	t->mask = (1U<<hash_bits) - 1;
	t->size = size;
	t->compare = compare;
	t->hash_table = calloc(size, sizeof(struct netl_hash_entry *));
	return (t->hash_table == NULL) ? -1 : 0;
}



/*
 * Insert a data with key into the table - 
 *  - already exists
 *  - out of memory
 */
static inline struct netl_hash_entry * 
netl_hash_insert(struct netl_hash_table *t, struct netl_hash_entry *e, uint32_t hash)
{
	int pos;
	struct netl_hash_entry *cur;
	struct netl_hash_entry **loc;
	
	pos = hash & t->mask;

	spin_lock(&t->hash_table_lock);
	loc = &t->hash_table[pos];
	cur = *loc;
	while (cur != NULL) {
		if (hash == cur->hash && (t->compare(cur, e) == 0)) {
			spin_unlock(&t->hash_table_lock);
			return cur;
		}
		loc = &(cur->next);
		cur = *loc;
	}

	e->hash = hash;
	e->next = NULL;

	*loc = e;

	spin_unlock(&t->hash_table_lock);
	return e;
	
}

/*
 */
static inline struct netl_hash_entry * 
netl_hash_lookup(struct netl_hash_table *t, struct netl_hash_entry *e, uint32_t hash)
{
	int pos;
	struct netl_hash_entry *cur;
	
	pos = hash & t->mask;

	spin_lock(&t->hash_table_lock);
	cur = t->hash_table[pos];
	while (cur != NULL) {
		if (hash == cur->hash && (t->compare(cur, e) == 0)) {
			spin_unlock(&t->hash_table_lock);
			return cur;
		}
		cur = cur->next;
	}

	spin_unlock(&t->hash_table_lock);
	return NULL;
}

static inline struct netl_hash_entry * 
netl_hash_remove(struct netl_hash_table *t, struct netl_hash_entry *e, uint32_t hash)
{

	int pos;
	struct netl_hash_entry *cur;
	struct netl_hash_entry **loc;

	pos = hash & t->mask;

	spin_lock(&t->hash_table_lock);
	loc = &t->hash_table[pos];
	cur = *loc;
	while (cur != NULL) {
		if (hash == cur->hash && (t->compare(cur, e) == 0))
			break;
		loc = &(cur->next);
		cur = *loc;
	}
	if (cur != NULL) {
		*loc = cur->next;
	}

	spin_unlock(&t->hash_table_lock);
	return cur;
}
#endif
