/***********************************************************************
Copyright 2003-2010 Netlogic Microsystems Inc. (“Netlogic”).
This is a derived work from software originally provided by the external
entity identified below. The licensing terms and warranties specified in
the header of the original work apply to this derived work.

*****************************#NETL_1#********************************/

/*-
 * Copyright (c) 1997 Doug Rabson
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are penlmtted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef _SYS_MODULE_H_
#define _SYS_MODULE_H_

#include "types.h"
#include "sys/sysproto.h"
#include "sys/linker.h"

/*
 * Module metadata types
 */
#define    MDT_DEPEND    1        /* argument is a module name */
#define    MDT_MODULE    2        /* module declaration */
#define    MDT_VERSION   3        /* module version(s) */

#define    MDT_STRUCT_VERSION    1    /* version of metadata structure */
#define    MDT_SETNAME    "modmetadata_set"

typedef enum modeventtype {
    MOD_LOAD,
    MOD_UNLOAD,
    MOD_SHUTDOWN,
    MOD_QUIESCE
} modeventtype_t;

typedef struct module *module_t;
typedef int (*modeventhand_t)(module_t, int /* modeventtype_t */, void *);

/*
 * Struct for registering modules statically via SYSINIT.
 */
typedef struct moduledata {
    const char      *name;        /* module name */
    modeventhand_t   evhand;      /* event handler */
    void            *priv;        /* extra data */
} moduledata_t;

/*
 * A module can use this to report module specific data to the user via
 * kldstat(2).
 */
typedef union modspecific {
    int     intval;
    u_int   uintval;
    long    longval;
    u_long  ulongval;
} modspecific_t;

/*
 * Module dependency declarartion
 */
struct mod_depend {
    int    md_ver_minimum;
    int    md_ver_preferred;
    int    md_ver_maximum;
};

/*
 * Module version declaration
 */
struct mod_version {
    int    mv_version;
};

struct mod_metadata {
    int          md_version; /* structure version MDTV_* */
    int          md_type;    /* type of entry MDT_* */
    void        *md_data;    /* specific data */
    const char  *md_cval;    /* common string label */
};

#include "sys/linker_set.h"
#include "sys/kernel.h"

#define    MODULE_METADATA(uniquifier, type, data, cval)            \
    static struct mod_metadata _mod_metadata##uniquifier = {    \
        MDT_STRUCT_VERSION,                    \
        type,                            \
        data,                            \
        cval                            \
    };                                \
    DATA_SET(modmetadata_set, _mod_metadata##uniquifier)

#define    MODULE_DEPEND(module, mdepend, vmin, vpref, vmax)        \
    static struct mod_depend _##module##_depend_on_##mdepend = {    \
        vmin,                            \
        vpref,                            \
        vmax                            \
    };                                \
    MODULE_METADATA(_md_##module##_on_##mdepend, MDT_DEPEND,    \
        &_##module##_depend_on_##mdepend, #mdepend)

#if 0
#define    DECLARE_MODULE(name, data, sub, order)                \
    MODULE_METADATA(_md_##name, MDT_MODULE, &data, #name);        \
    SYSINIT(name##module, sub, order, module_register_init, &data)    \
    struct __hack
#endif

#define    DECLARE_MODULE(name, data, sub, order)                \
    MODULE_METADATA(_md_##name, MDT_MODULE, &data, #name); \
    SYSINIT(name##module, sub, order, module_register_init, &data)

#define    MODULE_VERSION(module, version)                    \
    static struct mod_version _##module##_version = {        \
        version                            \
    };                                \
    MODULE_METADATA(_##module##_version, MDT_VERSION,        \
        &_##module##_version, #module)


struct linker_file;

void        module_register_init(const void *);
int         module_register(const struct moduledata *, struct linker_file *);
module_t    module_lookupbyname(const char *);
module_t    module_lookupbyid(int);
void        module_reference(module_t);
void        module_release(module_t);
int         module_unload(module_t, int flags);
int         module_getid(module_t);
module_t    module_getfnext(module_t);
void        module_setspecific(module_t, modspecific_t *);

#ifdef    MOD_DEBUG
extern int mod_debug;
#define    MOD_DEBUG_REFS    1

#define    MOD_DPF(cat, args) do {                        \
    if (mod_debug & MOD_DEBUG_##cat)                \
        printf(args);                        \
} while (0)

#else    /* !MOD_DEBUG */

#define    MOD_DPF(cat, args)
#endif


#define    MAXMODNAME    32

struct module_stat {
    int     version;    /* set to sizeof(struct module_stat) */
    char    name[MAXMODNAME];
    int     refs;
    int     id;
    modspecific_t    data;
};

extern int    kldload(const char* _file);
extern int    kldsym(struct kldsym_args *);
extern void   linker_init(void *arg);
extern int   link_elf_init(void* arg);
extern void  link_elf_obj_init(void* arg);
extern void   linker_stop_class_add(void *arg);
extern void   linker_init_kernel_modules(void);
extern void   module_init(void *arg);

static inline unsigned long callmodule(unsigned long function_addr, void *a)
{
    register unsigned long __faddr asm("$25") = function_addr;
    register unsigned long __a0 asm("$4") = (unsigned long) a;
    unsigned long __v0;

    __asm__ volatile (
        ".set  noreorder\n\t"
        "jalr  $25\n\t"
        "nop\n\t"
        "move  %0, $2\n\t"
        ".set  reorder\n\t"
        : "=&r" (__v0)
        : "r" (__a0), "r"(__faddr)
        : "$2", "$8", "$9", "$10", "$11", "$12", "$13", "$14", "$15", "$24","memory"
    );

    return __v0;
}

static inline unsigned long callmodule3(unsigned long function_addr,
                                        unsigned long a0,
                                        unsigned long a1,
                                        unsigned long a2)
{
    register unsigned long __faddr asm("$25") = function_addr;
    register unsigned long __a0 asm("$4") = a0;
    register unsigned long __a1 asm("$5") = a1;
    register unsigned long __a2 asm("$6") = a2;
    unsigned long __v0;

    __asm__ volatile (
        ".set  noreorder\n\t"
        "jalr  $25\n\t"
        "nop\n\t"
        "move  %0, $2\n\t"
        ".set  reorder\n\t"
        : "=&r" (__v0)
        : "r" (__a0), "r"(__a1), "r" (__a2), "r"(__faddr)
        : "$2", "$8", "$9", "$10", "$11", "$12", "$13", "$14", "$15", "$24","memory"
    );

    return __v0;
}

static inline unsigned long get_kld_symbol_addr(char *symbol_name)
{
    struct kldsym_args uap;
    struct kld_sym_lookup kld_arg;

    uap.fileid = 0;
    uap.cmd = 1;
    uap.data = (void *) &kld_arg;

    kld_arg.version = sizeof(struct kld_sym_lookup);
    kld_arg.symname = symbol_name;

    if (kldsym(&uap) != 0) {
        return 0;
    }

    return kld_arg.symvalue;
}

#define IMPORT_SYMBOL(sym)         \
    unsigned long __ksymtab_##sym  \
    __used                         \
    = (unsigned long)&sym

#endif    /* !_SYS_MODULE_H_ */
