/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <barrier_lib.h>
#include <irq.h>

#define dbg(a,b...) //printf("\n[LOADER--> %s @ %d] "a"\n",__FUNCTION__,__LINE__, ##b)
barrier_tag netl_barrier_head[MAX_BARRIER_HEAD];
static spinlock_t netl_barrier_lock;

static uint64_t calc_barrier_hash(const char *curr)
{
    uint64_t hash = 0;
    while(*curr){
        hash = hash + *curr;
        curr++;
    }
    return (hash%MAX_BARRIER_HEAD);
}

static barrier_tag *find_barrier(const char *name, barrier_tag **barrier_head)
{
    uint64_t hash = 0;
    barrier_tag *head = NULL;
    barrier_tag *entry = NULL;
    struct list *addr = NULL;

    /*calculate hash*/
    hash = calc_barrier_hash(name);
    head = &netl_barrier_head[hash];
    *barrier_head= head;

    list_iterate_forward(&head->list, addr){
        entry  = list_container_of(barrier_tag, addr, list);
        if(!strcmp(entry->name, name))
            return entry;
    }
    return NULL;
}

static barrier * _barrier_open_r(const char *name, int oflag, int weight)
{
    unsigned long flags;
    barrier_tag *head = NULL;
    barrier_tag *entry = NULL;

    if((oflag & ~(O_CREAT|O_EXCL)) || (oflag == O_EXCL))
        return BARRIER_FAILED;

    spin_lock_irqsave(&netl_barrier_lock, flags);
    entry =  find_barrier(name, &head);

    if(!entry){
        /*Allocate new barrier entry and add it to the hash list*/

        if(weight > BARRIER_VALUE_MAX)
            goto error;

        entry = (barrier_tag *)malloc(sizeof(barrier_tag));
        if(!entry)
            goto error;

        entry->name = malloc(strlen(name)+1);
        if(!entry->name){
            free(entry);
            goto error;
        }
        strcpy(entry->name, name);
        entry->magic = BARRIER_MAGIC;
        entry->weight = weight;
        entry->oflag = oflag;
        entry->flag =  0;
        entry->users  = 1;
        entry->unlink = 0;
        spin_lock_init(&entry->lock);
        list_add_tail(&head->list, &entry->list);
    }else{
        if(entry->magic != BARRIER_MAGIC)
            goto error;
        if(oflag & O_EXCL){
            /*return error if oflag has O_EXCL attribute*/
            goto error;
        }
        entry->users++;
    }
    spin_unlock_irqrestore(&netl_barrier_lock, flags);
    return &entry->weight;
error:
    spin_unlock_irqrestore(&netl_barrier_lock, flags);
    return BARRIER_FAILED;
}

static int _barrier_close_r(barrier_tag *barrier)
{
    unsigned long flags;

    if(barrier->magic != BARRIER_MAGIC)
        return -1;

    spin_lock_irqsave(&netl_barrier_lock, flags);
    barrier->users--;
    if(!barrier->users && barrier->unlink){
        free(barrier->name);
        free(barrier);
    }
    spin_unlock_irqrestore(&netl_barrier_lock, flags);
    return 0;
}

static int _barrier_unlink_r(const char *name)
{
    unsigned long flags;
    barrier_tag *head = NULL;
    barrier_tag *entry = NULL;

    spin_lock_irqsave(&netl_barrier_lock, flags);

    entry =  find_barrier(name, &head);

    if(!entry || entry->magic != BARRIER_MAGIC){
        spin_unlock_irqrestore(&netl_barrier_lock, flags);
        return -1;
    }

    list_delete(&entry->list);
    if(entry->users == 0){
        /*free all system resources..*/
        free(entry->name);            
        free(entry);
    }else{
        entry->unlink = 1;
    }
    spin_unlock_irqrestore(&netl_barrier_lock, flags);
    return 0;
}

int barrier_sys_call(struct barrier_param *param)
{
    int ret = 0;
    switch(param->index)
    {
        case BARRIER_OPEN:
            dbg("Open called!!");
            param->open.ret = _barrier_open_r(param->open.name, 
                            param->open.oflag, param->open.weight);
            break;
        case BARRIER_CLOSE:
            dbg("Close called!!");
            param->close.ret = _barrier_close_r(param->close.barrier);
            break;
        case BARRIER_UNLINK:
            dbg("Unlink called!!");
            param->unlink.ret = _barrier_unlink_r(param->unlink.name);
            break;
        default:
            ret = -1;
            break;    
    }
    return ret;
}

/*Init barrier data structures..*/
void barrier_init()
{
    int i = 0;
    for(i=0; i<MAX_BARRIER_HEAD; i++){
        memset(&netl_barrier_head[i], 0, sizeof(barrier_tag));
        list_init(&netl_barrier_head[i].list);
    }
}
