//
// readline.c
//
// Read line with editing
//
// Copyright (C) 2002 Michael Ringgaard. All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are penlmtted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright 
//    notice, this list of conditions and the following disclaimer.  
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.  
// 3. Neither the name of the project nor the names of its contributors
//    may be used to endorse or promote products derived from this software
//    without specific prior written penlmssion. 
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
// SUCH DAMAGE.
// 

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

#include <uart.h>
#include "stubs.h"

#define MAX_HISTORY     64

#define KEY_BACKSPACE   0x08
#define KEY_ENTER       0x0D
#define KEY_ENTER_SIM   0xa
#define KEY_ESC         0x102
#define KEY_INS         0x103
#define KEY_DEL         0x104
#define KEY_LEFT        0x105
#define KEY_RIGHT       0x106
#define KEY_UP          0x107
#define KEY_DOWN        0x108
#define KEY_KILL        0x109
#define KEY_YANK        0x10A
#define KEY_HOME        0x10B
#define KEY_END         0x10C
#define KEY_TAB         0x10E
#define KEY_CTRL_C      0x10F
#define KEY_HIST        0x110
#define KEY_LAST        0x111

#define KEY_UNKNOWN     0xFFF

#define CMD_BUF_SIZE 256


static int insmode = 1;
static int hist_idx = 0;
static int history_len = 0;

static char *history[MAX_HISTORY];

void add_to_history(char *line)
{
	int i;

	if (!line || !*line) return;
	if (history_len > 0 && strcmp(line, history[history_len - 1]) == 0) return;

	if (history_len == MAX_HISTORY)
	{
		free(history[0]);
		for (i = 0; i < history_len - 1; i++) history[i] = history[i + 1];
		history_len--;
	}

	history[history_len] = malloc(strlen(line) + 1);
	if (history[history_len]) 
	{
		strcpy(history[history_len], line);
		history_len++;
	}
}

static int getkey()
{
	int ch;

	ch = inbyte();

	if (ch <= 0) 
	{
		return -1;
	}
	switch (ch)
	{
	case 0x01: /* Ctrl-A */
		return KEY_HOME;
	case 0x02: /* Ctrl-B */
		return KEY_LEFT;
	case 0x03: /* Ctrl-C */
		return KEY_CTRL_C;
	case 0x04: /* Ctrl-D */
		return KEY_DEL;
	case 0x05: /* Ctrl-E */
		return KEY_END;
	case 0x10: /* Ctrl-P */
		return KEY_UP;
	case 0x0E: /* Ctrl-N */
		return KEY_DOWN;
	case 0x06: /* Ctrl-F */
		return KEY_RIGHT;
	case 0x0B: /* Ctrl-K */
		return KEY_KILL;
	case 0x19: /* Ctrl-Y */
		return KEY_YANK;

	case 0x08: 
	case 0x7f:
		return KEY_BACKSPACE;
	case 0x09: 
		return KEY_TAB;

	case 0x1B:
		ch = inbyte();
		switch (ch)
		{
		case 0x1B: 
			return KEY_ESC;
		case 0x5B:
			ch = inbyte();
			switch (ch)
			{
			case 0x31:
				ch = inbyte();
				switch (ch)
				{
				case 0x7E:
					return KEY_HOME;
				default:
					return KEY_UNKNOWN;
				}
				break;
			case 0x32:
				ch = inbyte();
				switch (ch)
				{
				case 0x7E:
					return KEY_INS;
				default:
					return KEY_UNKNOWN;
				}
				break;
			case 0x33:
				ch = inbyte();
				switch (ch)
				{
				case 0x7E:
					return KEY_DEL;
				default:
					return KEY_UNKNOWN;
				}
				break;
			case 0x34:
				ch = inbyte();
				switch (ch)
				{
				case 0x7E:
					return KEY_END;
				default:
					return KEY_UNKNOWN;
				}
				break;
			case 0x41: 
				return KEY_UP;
			case 0x42: 
				return KEY_DOWN;
			case 0x43: 
				return KEY_RIGHT;
			case 0x44: 
				return KEY_LEFT;
			default: 
				return KEY_UNKNOWN;
			}
			break;

		default:
			return KEY_UNKNOWN;
		}

		break;

	default: 
		return ch;
	}
}

int readline(char *buf, int bufsize)
{
	static int key = 0;
	static int i   = 0;
	static char kill_buf[CMD_BUF_SIZE];
	static int kill_len;
	static int idx = 0;
	static int len = 0;
	int done = 0;

	if (bufsize <= 0) 
	{
		errno = EINVAL;
		return -1;
	}

	fflush(stdout);
	key = getkey();
	if (key < 0) 
	{
		errno = -key;
		return -1;
	}

	switch (key)
	{
	case KEY_LEFT:
		if (idx > 0)
		{
			putchar('\b');
			idx--;
		}
		break;

	case KEY_RIGHT:
		if (idx < len)
		{
			putchar(buf[idx]);
			idx++;
		}
		break;
	case KEY_DEL:
		if (idx < len)
		{
			len--;
			memmove(buf + idx, buf + idx + 1, len - idx);
			for (i = idx; i < len; i++) putchar(buf[i]);
			putchar(' ');
			putchar('\b');
			for (i = idx; i < len; i++) putchar('\b');
		}
		break;

	case KEY_KILL:
		if (idx < len)
		{
			kill_len = len - idx;
			memset(kill_buf,0,sizeof(kill_buf));
			memcpy(kill_buf,buf+idx,kill_len);
			while (len > idx)
			{
				putchar(' ');
				len--;
			}
			for(i=0; i<kill_len;i++)
				putchar('\b');
		}
		break;
	case KEY_YANK:
		if((len+kill_len) < (bufsize-1))
		{
			if (idx < len) memmove(buf+idx+kill_len, buf+idx, len-idx);
			for(i=0; i<kill_len; i++)
			{
				buf[idx] = kill_buf[i];
				putchar(buf[idx]);
				idx++;len++;
			}
			for(i=idx;i<len;i++) putchar(buf[i]);
			for(i=idx;i<len;i++) putchar('\b');
		}
		break;
	case KEY_INS:
		insmode = !insmode;
		break;
	case KEY_HOME:
		while (idx > 0)
		{
			putchar('\b');
			idx--;
		}
		break;

	case KEY_END:
		while (idx < len)
		{
			putchar(buf[idx]);
			idx++;
		}
		break;

	case KEY_BACKSPACE:
		if (idx > 0)
		{
			putchar('\b');
			idx--;
			len--;
			memmove(buf + idx, buf + idx + 1, len - idx);
			for (i = idx; i < len; i++) putchar(buf[i]);
			putchar(' ');
			putchar('\b');
			for (i = idx; i < len; i++) putchar('\b');
		}
		break;

	case KEY_ENTER :
	#if defined(PHOENIX_SIM) || defined(XLP_SIM)
	case KEY_ENTER_SIM:
	#endif
		putchar('\r');
		putchar('\n');
		idx = 0;
		key = 0;
		i   = 0;
		done = 1;
		break;

	case KEY_CTRL_C:
		putchar('\r');
		putchar('\n');
		idx = 0;
		key = 0;
		i   = 0;
		len = 0;
		done =1;
		break;

	case KEY_UP:
		if (hist_idx > 0)
		{
			hist_idx--;
			for (i = 0; i < idx; i++) putchar('\b');
			for (i = 0; i < len; i++) putchar(' ');
			for (i = 0; i < len; i++) putchar('\b');
			len = strlen(history[hist_idx]);
			if (len > bufsize - 1) len = bufsize - 1;
			idx = len;
			memcpy(buf, history[hist_idx], len);
			for (i = 0; i < len; i++) putchar(buf[i]);
		}
		break;

	case KEY_DOWN:
		if (hist_idx < history_len - 1)
		{
			hist_idx++;
			for (i = 0; i < idx; i++) putchar('\b');
			for (i = 0; i < len; i++) putchar(' ');
			for (i = 0; i < len; i++) putchar('\b');
			len = strlen(history[hist_idx]);
			if (len > bufsize - 1) len = bufsize - 1;
			idx = len;
			memcpy(buf, history[hist_idx], len);
			for (i = 0; i < len; i++) putchar(buf[i]);
		}
		break;

	case KEY_LAST:
		len = strlen(history[hist_idx-1]);
		memcpy(buf, history[hist_idx-1], len);
		break;
		
	case KEY_UNKNOWN:
		break;

	default:
		if (key >= 0x20 && key <= 0xFF)
		{
					
			if (insmode)
			{
				if (len < bufsize - 1)
				{
					if (idx < len) memmove(buf + idx + 1, buf + idx, len - idx);
					buf[idx] = key;
					len++;
					for (i = idx; i < len; i++) putchar(buf[i]);
					idx++;
					for (i = idx; i < len; i++) putchar('\b');
				}
			}
			else
			{
				if (idx < bufsize - 1)
				{
					buf[idx] = key;
					putchar(buf[idx]);
					if (idx == len) len++;
					idx++;
				}
			}
		}
		break;
	}

	if(done)
	{
		int tlen;
		if(buf[0] == '!')
		{
			if(buf[1] == '!')
			{
				if(history_len)
				{
					len = strlen(history[history_len-1]);
					memcpy(buf,history[history_len-1],len);
					printf("%s\n",buf);
				}
				else
				{
					len = 0;
				}
			}
			else
			{
				int hnum = atoi(&buf[1]);
				if(history[hnum])
				{
					len = strlen(history[hnum]);
					memcpy(buf,history[hnum],len);
					printf("%s\n",buf);
				}
				else
				{
					len = 0;
				}
			}				
		}
		buf[len] = 0;
		tlen = len +1;
		len = 0;
		add_to_history(buf);
		hist_idx = history_len;
		return tlen;
	}
	else
	{
		return 0;
	}
}
	
static void print_history(int argc, char *argv[])
{
	int i;
	for(i=0 ; i <MAX_HISTORY; i++)
	{
		if(history[i])
			printf("%d %s\n",i, history[i]);
	}
}

void history_init(int nr_hist)
{
	register_shell_cmd("hist",print_history, "Print History");
}
