/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#include <semlib.h>
#include <string.h>
#include <stdio.h>

#define dbg(a,b...) //printf("\n[LOADER--> %s @ %d] "a"\n",__FUNCTION__,__LINE__, ##b)
struct semaphore netl_sema_head[MAX_SEM_HEAD];
static spinlock_t netl_sem_lock;

static uint64_t calc_sem_hash(const char *curr)
{
    uint64_t hash = 0;
    while(*curr){
        hash = hash + *curr;
        curr++;
    }
    return (hash%MAX_SEM_HEAD);
}

static struct semaphore *find_semaphore(const char *name, struct semaphore **sem_head)
{
    uint64_t hash = 0;
    struct semaphore *head = NULL;
    struct semaphore *entry = NULL;
    struct list *addr = NULL;

    /*calculate hash*/
    hash = calc_sem_hash(name);
    head = &netl_sema_head[hash];
    *sem_head = head;

    list_iterate_forward(&head->list, addr){
        entry  = list_container_of(struct semaphore, addr, list);
        if(!strcmp(entry->name, name))
            return entry;
    }
    return NULL;
}

static sem_t *_sem_open_r(const char *name, int oflag, mode_t mode, unsigned int value)
{
    struct semaphore *head = NULL;
    struct semaphore *entry = NULL;
    unsigned long flags;

    if((oflag & ~(O_CREAT|O_EXCL)) || (oflag == O_EXCL)){
        return SEM_FAILED;
    }

    spin_lock_irqsave(&netl_sem_lock, flags);

    entry =  find_semaphore(name, &head);

    if(!entry){
        /*Allocate new semaphore entry and add it to the hash list*/

        if(value > SEM_VALUE_MAX)
            goto error;

        entry = (struct semaphore *)malloc(sizeof(struct semaphore));
        if(!entry)
            goto error;

        entry->name = malloc(strlen(name)+1);
        if(!entry->name){
            free(entry);
            goto error;
        }

        strcpy(entry->name, name);
        entry->sem = value;
        entry->users = 1;
        entry->oflag = oflag;
        entry->magic = SEM_MAGIC;
        entry->unlinked = 0;
        list_add_tail(&head->list, &entry->list);
    }else{
        if(entry->magic != SEM_MAGIC)
            goto error;
        if(oflag & O_EXCL){
            /*return error if oflag has O_EXCL attribute*/
            goto error;
        }
        if(entry->users == SEM_USER_MAX)
            goto error;
        entry->users++;
    }

    spin_unlock_irqrestore(&netl_sem_lock, flags);
    return &entry->sem;

error:
    spin_unlock_irqrestore(&netl_sem_lock, flags);
    return SEM_FAILED;
}

static int _sem_close_r(sem_t *sem)
{
    struct semaphore *entry= SEM_TO_SEMAPHORE(sem);
    unsigned long flags;

    if(entry->magic != SEM_MAGIC)
        return -1;

    spin_lock_irqsave(&netl_sem_lock, flags);
    if(entry->unlinked){
        /*unlink is already done on this semaphore.*/
        if(entry->users == 1){
            /*last user.. free all system resources*/
            free(entry->name);
            free(entry);
            spin_unlock_irqrestore(&netl_sem_lock, flags);
            return 0;
        }
    }
    entry->users--;
    spin_unlock_irqrestore(&netl_sem_lock, flags);
    return 0;
}

static int _sem_unlink_r(const char *name)
{
    struct semaphore *head, *entry;
    unsigned long flags;

    spin_lock_irqsave(&netl_sem_lock, flags);

    entry = find_semaphore(name, &head);

    if(!entry || (entry->magic != SEM_MAGIC)){
        spin_unlock_irqrestore(&netl_sem_lock, flags);
        return -1;
    }

    /*remove the entry..*/
    list_delete(&entry->list);
    if(entry->users == 0){
        /*free all system resources..*/
        free(entry->name);
        free(entry);
    }else{
        entry->unlinked = 1;
    }

    spin_unlock_irqrestore(&netl_sem_lock, flags);
    return 0;
}

/*Init semaphore data structures..*/
void semaphore_init()
{
    int i = 0;
    dbg("called...\n");
    for(i=0; i<MAX_SEM_HEAD; i++){
        memset(&netl_sema_head[i], 0, sizeof(struct semaphore));
        list_init(&netl_sema_head[i].list);
    }
}

int sem_sys_call(void *data)
{
    struct semparam *param = (struct semparam *)data;
    int ret = 0;
    dbg("index = %d",param->index);
    switch(param->index)
    {
        case SEM_OPEN:
            param->open.ret = _sem_open_r(param->open.name, param->open.oflag,
                                        param->open.mode, param->open.value);
            break;
        case SEM_CLOSE:
            param->close.ret = _sem_close_r(param->close.sem);
            break;
        case SEM_UNLINK:
            param->unlink.ret = _sem_unlink_r(param->unlink.name);
            break;
        default:
            ret = -1;
            break;
    }
    return ret;
}

#ifdef SEM_DEBUG
static void dump_sem_entry(int index, int semindx, struct semaphore *sem)
{
    printf("**********HEAD_INDEX = %d --  SEM_INDEX = %d**********************\n",index,semindx);
    printf("Name = %s, SemValue = %#x, SemPtr = %p, Magic = %#x\n Users = %d, Oflag = %#x, Unlinked = %d\n", sem->name, sem->sem, &sem->sem, sem->magic, sem->users, sem->oflag, sem->unlinked);
}

void dump_semaphore()
{
    struct semaphore *head = NULL;
    struct semaphore *entry = NULL;
    struct list *addr = NULL;
    int i=0, j = 0;
    unsigned long flags;

    spin_lock_irqsave(&netl_sem_lock, flags);
    printf("-----DUMP STARTS-----\n");
    for(i=0; i<MAX_SEM_HEAD; i++){
        head = &netl_sema_head[i];
        j = 0;
        list_iterate_forward(&head->list, addr){
            entry  = list_container_of(struct semaphore, addr, list);
            dump_sem_entry(i, j++, entry);
        }
    }
    printf("----DUMP ENDS-----\n");
    spin_unlock_irqrestore(&netl_sem_lock, flags);
    return;
}
#endif
