#include <stdio.h>
#include <string.h>

#include <uart.h>
#include "stubs.h"
#ifndef XLP_SIM
#include "iomap.h"
#endif

/* VT102 emulation */
#define DEL    0x7f
#define BEL    0x07
#define BS     0x08
#define ETX    0x03
#define CTRL_C ETX
#define NAK    0x15
#define CTRL_U NAK
#define ETB    0x17
#define CTRL_W ETB
#define SOH    1
#define CTRL_A SOH
#define VT     0x0b
#define CTRL_K VT
#define TAB    0x09

#define MAX_ARGV_ARGS 32
#define MAX_CMDS 100
#define CMD_NAME_SIZE 32
#define CMD_SHHELP_SIZE 64

#ifdef XLR
extern unsigned long phoenix_io_base
#endif
extern int readline(char *buf, int bufsize);
extern void history_init(int nr_hist);

static char shell_cmd_buf[256];
static char *argv[MAX_ARGV_ARGS];

struct cmd {
	char name[CMD_NAME_SIZE];
	char shorthelp[CMD_SHHELP_SIZE];
	void (*func)(int, char *[]);
};

static struct cmd cmd_list[MAX_CMDS];

#define DEFAULT_PROMPT "NetOS $ "

char prompt[32];

void execute_cmd(char *buf)
{
	const char    *delim = " \t";
	int            argc = 0, i = 0;

	/* Parse command line and setup args 
	 */
	argv[argc++] = strtok(buf, delim);
	if ( !argv[argc-1] ) {
		printf("Bad command token??\n");
		return;
	}

	while(1) {
		/* Can't handle more than 10 arguments */
		if (argc >= MAX_ARGV_ARGS) 
			break;

		argv[argc++] = strtok(0, delim);
		if (!argv[argc-1]) {
			argc--;
			break;
		}
	}

	/* Search for the command */
	for (i = 0; i < MAX_CMDS; i++) {
		if (strcmp(cmd_list[i].name, argv[0]) == 0) 
			break;
	}
	if (i >= MAX_CMDS) {
		printf("%s: command not found\n", buf);
		return;
	}
	
	if (!cmd_list[i].func) {
		printf("Bad function pointer!\n");
		return;
	}

	/* call the command handler */
	(cmd_list[i].func)(argc, argv);

}

void print_prompt(void)
{
	printf("\n");
	printf(prompt);
	fflush(stdout);
}

void shell_run(void)
{
	int len = 0;

	len = readline(shell_cmd_buf, sizeof(shell_cmd_buf));

	if(len <= 0)
	{
		return;
	}
	else
	{
		if(len > 1)
		{
			execute_cmd(shell_cmd_buf);
		}
		print_prompt();
	}
}

int shell_app_init(void)
{
	print_prompt();
	printf("\n");
	print_prompt();
	return 0;
}

__attribute__((noreturn)) void shell(void)
{
	int    ch;
	int    i = 0;

	print_prompt();
	while (1) {
		ch = getchar();
		putchar(ch);    
		shell_cmd_buf[i++] = ch;

		if (ch == '\n') {
			shell_cmd_buf[i-1] = 0;
			if (shell_cmd_buf[0] != 0)
				execute_cmd(shell_cmd_buf);
			i = 0;
			print_prompt();
		}
		else if (ch == DEL) {
			if (i != 0) 
				i--;
			else 
				putchar(BEL);
		}
	}
}

int register_shell_cmd(const char *name, void (*func)(int, char *[]),  
		const char *shorthelp)
{
	int i = 0, ret = -1;

	if (!name || !name[0]) 
		goto out;

	for (i = 0; i < MAX_CMDS && cmd_list[i].name[0]; i++) 
		;
	if (i >= MAX_CMDS) 
		goto out;

	strncpy(cmd_list[i].name, name, CMD_NAME_SIZE );
	strncpy(cmd_list[i].shorthelp, shorthelp, CMD_SHHELP_SIZE);
	cmd_list[i].func = func;

	ret = 0;
out:
	if (ret < 0) {
		printf("Failed to register shell cmd [%s]\n", name);
	}
	return ret;
}

int unregister_shell_cmd(const char *name)
{
	int    i = 0;
	int    ret = -1;
	
	if (!name || !name[0]) 
		goto out;
	
	for (i = 0; i < MAX_CMDS && strcmp(name, cmd_list[i].name) != 0; i++) 
		;
	if (i == MAX_CMDS) 
		goto out;
	
	cmd_list[i].name[0] = 0;
	
	ret = 0;
 out:
	printf("Failed to unregister shell cmd [%s]\n", name);
	return ret;
}

void shell_help(int argc, char *argv[])
{
	int i = 0;
	char a;

	printf("Available Commands: \n\n");

	for (i = 0; i < MAX_CMDS; i++) {
		if (cmd_list[i].name[0]) {
			printf("%-28s", cmd_list[i].name);
			printf(" - %s\n", cmd_list[i].shorthelp);
		}
		else {
			continue;
		}

		if ((i != 0) && ((i % 30) == 0)) {
			printf( " Press Enter to Continue or  Q to Quit  \n");
			a = inbyte();
			switch (a) {
				case 0x0d: {
				   if ( i >= MAX_CMDS )
					   return ;
				   }
				   break;
				case 'q':
				case 'Q':
				   return ;
			}
		}
	}
	printf("\n");
}

void change_prompt(int argc, char *argv[])
{
	if (argc < 2 || !argv[1]) {
		printf("Bad args, argc = %d, argv[1] = [%s]\n", argc, argv[1]);
		return;
	}
	strncpy(prompt, argv[1], 32);
}

#ifdef XLP_SIM
#include "nlm_hal.h"
static void chip_reboot(int argc, char *argv[])
{
	printf("===================================\n");
	printf("Resetting the chip...\n");
	printf("===================================\n");

	nlm_hal_write_32bit_reg(xlp_sys_base, 0x40, 1);
}
#else
#define GPIO_RESET  8
static void chip_reboot(int argc, char *argv[])
{
	phoenix_reg_t *mmio = (phoenix_reg_t *)(phoenix_io_base + PHOENIX_IO_GPIO_OFFSET);

	printf("===================================\n");
	printf("Resetting the chip...\n");
	printf("===================================\n");
	phoenix_write_reg(mmio, GPIO_RESET, 1);
}
#endif

void shell_init(void)
{
	int i=0;

	strncpy(prompt, DEFAULT_PROMPT, 32);
	for (i = 0; i < MAX_CMDS; i++)
		cmd_list[i].name[0] = 0;

	history_init(64);

	register_shell_cmd("help", shell_help, "Help");
	register_shell_cmd("h", shell_help, "Help");
	register_shell_cmd("reboot", chip_reboot, "Reboot Chip");
}
