/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <fcntl.h>
#include "spinlock.h"
#include "hash.h"
#include "shmparam.h"
#include "contig_alloc.h"

#define netl_roundup(x, y)   ((((x)+((y)-1))/(y))*(y))
#define S_IRWUSR 0600

static spinlock_t shminfo_lock;

static struct netl_hash_table shm_open_hash;
static int shm_open_compare(struct netl_hash_entry *cur, struct netl_hash_entry *entry)
{
	struct shm_open_entry *ocur, *oentry;

	ocur =   ((struct shm_open_entry *)((void *)(cur) - offsetof(struct shm_open_entry, sentry)));
	oentry = ((struct shm_open_entry *)((void *)(entry) - offsetof(struct shm_open_entry, sentry)));
	if((strcmp(ocur->name, oentry->name) == 0)) {
		return 0;
	}
	return -1;
}

/*
 * DJB's string hash - good enough.
 */
static uint32_t hash_string(const char *str, int len)
{
	uint32_t hash = 5381;
	uint8_t c;

	while (len--) {
		c = *str++;
		hash = ((hash << 5) + hash) + c; /* hash * 33 + c */
	}

	return hash;
}

void *netl_shm_open(const char *name, int oflag, int mode, int *err)
{
	struct shm_open_entry *shm = NULL, *rshm = NULL; 
	struct netl_hash_entry *rh = NULL;
	uint32_t hash;

	if((shm = calloc(1, sizeof(*shm))) == NULL) {
		*err = ENOSPC;
		return NULL;
	}
	strncpy(shm->name, name, PATH_MAX);
	hash = hash_string(name, strlen(name));


	if(oflag & O_CREAT) {
		/* Call with lock held */
		spin_lock(&shminfo_lock); 

		if((oflag & O_RDWR) != O_RDWR) {
			fprintf(stderr, "Error in oflag: We can create only shm mem segs with rw permissions\n");
			*err = EACCES;
			goto err_exit;
		}
		if((mode & S_IRWUSR) != S_IRWUSR) {
			fprintf(stderr, "Error in mode: We can create only shm mem segs with rw permissions\n");
			*err = EACCES;
			goto err_exit;
		}
		
		rh = netl_hash_insert(&shm_open_hash, &shm->sentry, hash);
		if(rh == NULL) {
			*err = ENOSPC;
			goto err_exit;
		}
		if(rh == &shm->sentry) { /* if this is a new entry */
			shm->len = 0;
			shm->mem = NULL;
			shm->refcnt++;
			spin_unlock(&shminfo_lock);
			return shm;
		}

		/* if this is an already existing entry , return errorn if excl flag set. 
		truncate to zero if trunc flag set */
		rshm = ((struct shm_open_entry *)((void *)(rh) - offsetof(struct shm_open_entry, sentry)));
		if(oflag & O_EXCL) {
			*err = EEXIST;
			goto err_exit;
		}
		if(oflag & O_TRUNC) {
			spin_lock(&rshm->lock);
			if(rshm->mem != NULL)
				contig_free(rshm->mem);
			rshm->mem = NULL;
			rshm->len = 0;
			spin_unlock(&rshm->lock);
		}
		free(shm);
		rshm->refcnt++;
		spin_unlock(&shminfo_lock); 
		return rshm;
	} else {
		spin_lock(&shminfo_lock); 
		rh = netl_hash_lookup(&shm_open_hash, &shm->sentry, hash);
		if(rh == NULL) {
			*err = ENOENT;
			goto err_exit;
		}
		rshm = ((struct shm_open_entry *)((void *)(rh) - offsetof(struct shm_open_entry, sentry)));
		free(shm);
		rshm->refcnt++;
		/* give the lock taken by compare func */
		spin_unlock(&shminfo_lock); 
		return rshm;
	}

err_exit:
	if(shm)
		free(shm);

	spin_unlock(&shminfo_lock); 
	return NULL;
}

int netl_shm_destroy(struct shm_open_entry *shm)
{
	spin_lock(&shminfo_lock);

	if(shm->refcnt == 0) {
		printf("destroy name = %s \n",shm->name);
		if(shm->mem)
			contig_free(shm->mem);
		free(shm);
	}
	spin_unlock(&shminfo_lock);
	return 0;
}



int netl_shm_unlink(const char *name, int *err)
{
	struct shm_open_entry *shm = NULL;
	struct netl_hash_entry *rh = NULL;
	uint32_t hash;

	if((shm = calloc(1, sizeof(*shm))) == NULL) {
		*err = ENOSPC;
		return -1;
	}
	strncpy(shm->name, name, PATH_MAX);
	hash = hash_string(name, strlen(name));

	spin_lock(&shminfo_lock);

	rh = netl_hash_remove(&shm_open_hash, &shm->sentry, hash);
	if(rh == NULL) {
		free(shm);
		*err = ENOENT;	
		spin_unlock(&shminfo_lock);
		return -1;
	}
	free(shm);

	shm = ((struct shm_open_entry *)((void *)(rh) - offsetof(struct shm_open_entry, sentry)));
	shm->unlink = 1;

	if(shm->refcnt == 0) {
		if(shm->mem)
			contig_free(shm->mem);
		free(shm);
	} 

	spin_unlock(&shminfo_lock);
	return 0;
}


int netl_shm_init(void)
{
	int rc;
	/* 1k of hash */
	if((rc = netl_hash_table_init(&shm_open_hash, 10, shm_open_compare)) != 0)
		return rc;
	return 0;
		
}

void netl_dump_hash_info()
{
	struct netl_hash_table *t;
        struct netl_hash_entry *n;
	int i;

	printf(" shm open hash\n");
	t = &shm_open_hash;
	for(i=0; i < t->size ; i++) {
                for (n = t->hash_table[i]; n; n=n->next) {
			struct shm_open_entry *fe;
			fe =   ((struct shm_open_entry *)((void *)(n) - offsetof(struct shm_open_entry, sentry)));
			printf(" pos=%d(0x%x) hash=%x name=%s mem=%lx refcnt %d unlink %d\n", 
				i, i, n->hash, fe->name, (long)fe->mem, fe->refcnt, fe->unlink);
		}
	}
}

int shm_sys_call(void *data)
{
	struct netl_shmcall *m = data;
	//printf("shm_sys_call cmd = %d\n", m->cmd);
	switch(m->cmd)
	{
		case NETL_SHMOPEN_CALL:
		{
			struct netl_shmopen *open = &m->shmopen;
			open->obj = (long)netl_shm_open(open->name, open->oflag, open->mode, &m->err);
			break;
		}
		case NETL_SHMDESTROY_CALL:
		{
			struct netl_shmdestroy *destroy = &m->shmdestroy;
			m->rv = netl_shm_destroy((void *)(long)destroy->obj);
			break;
		}
		case NETL_SHMUNLINK_CALL:
		{
			struct netl_shmunlink *unlink = &m->shmunlink;
			m->rv = netl_shm_unlink(unlink->name, &m->err);
			break;
		}
		default:
			m->rv = -1;
			return -1;
	}
	m->rv = 0;
	return 0;
}
