/*************************************************************************
  Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
  reserved.
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the
  distribution.
  THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
  THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#include <unistd.h>
#include <stdio.h>
#include <string.h>

#include "xlr_boot_lib.h"
#include "elf.h"
#include "elfstructs.h"
#include "libelf.h"
#include "elf-ext.h"
#include "memblk.h"

#define BIT_SHARED 3
#define nlm_roundup(x, y)   (nlm_align((x)+(y)-1, (y)))
#define nlm_align(x, y)     ((x) & (~((y)-1)))

/* Global Variables */
uint64_t heap_size = 0, contig_alloc_off = 0, contig_alloc_sz = 0;
static struct ext_tlb curr_tlb_mapping;
unsigned int override_checks = 1;
static int commoncode_seg_enabled = 1;
unsigned long long app_endaddr;

/* Extern variables */
extern int lib_verbose;
extern int nseg;
extern Elf_ExtPhdr segment_map[MAX_ELF_SEGMENTS];
extern unsigned int kseg_master_cpu;
extern lib_spinlock_t print_lock;
extern int kseg2_image;	/* shared kseg2-kseg0 image, needed for MMU enabled vxWorks */
extern uint64_t xlr_lib_kseg0_start, xlr_lib_kseg0_size;
extern unsigned long kuseg_master_cpu_mask;
extern volatile unsigned long success_mask;
extern volatile struct xlr_vcpu_wakeup_info *xlr_vcpu_wakeup_info; 

/* Forward Declerations */
static int load_elf_segment(struct elf_binary *elf, Elf_ExtPhdr * segment,
							unsigned long segoff, size_t size,
							uint64_t paddr, int lib_run_mode,
							int loading_nmi_image);

/* Extern Function Declerations */
extern void free_tlb_list_nomem(int flag);
extern int get_mapped_tlb_num();
extern struct ext_tlb *combine_tlb_list();
extern int build_tlb_list(struct ext_tlb *tlb, int seg_num);
extern int check_tlb_condition(struct ext_tlb *tlb_ptr, int num_tlbs);
extern void release_exclusive_tlb_list();
extern void fill_per_cpu_tlb_info(struct ext_tlb *tlb_ptr, int cpu_index, int num_tlbs);
extern void cleanup_tlb_list();
extern void memcpy64_phys_phys(uint64_t dst, uint64_t src, size_t n);
extern void __memset_64(uint64_t dst, int c, size_t n);
extern void memcpy_virt_phys(unsigned char *dest, uint64_t src, int nbytes);

/************************************************************************************/
static void
print_segment_type(unsigned int flags, int is_shared)
{
	if (flags & PF_SHR) {
		printf("loadinfo:  shared data segment ");
	} else if (flags & PF_X) {
		printf("loadinfo:  code segment ");
	} else {
		printf("loadinfo:  data segment ");
	}

	if (is_shared) {
		printf("(shared):\n");
	} else {
		printf("\n");
	}
}

static void
print_segment_map(struct elf_binary *elf, int cpu_id)
{
	int i, j;
	Elf_ExtPhdr *segment;

	printf("loadinfo:  ------ memory pap for cpu %d ------\n", cpu_id);

	for (i = 0; i < nseg; i++) {
		segment = &segment_map[i];
		print_segment_type(elf_uval(elf, &(segment->hdr), p_flags),
						   segment->is_shared);

		for (j = 0; j < segment->ntlb; ++j) {
			printf
				("loadinfo:  vaddr: 0x%llx  paddr: 0x%llx  tlb_size: 0x%x\n",
				 (long long) (unsigned long) segment->
				 map[j].virt | ((long long) segment->map[j].asid << 32),
				 (unsigned long long) segment->map[j].phys,
				 segment->map[j].page_size);
		}
	}
	printf("\n");
}

void
copy_data(uint64_t dst, uint64_t src, unsigned int n)
{
	memcpy64_phys_phys(dst, src, n);
}

void
set_data(unsigned long long dst, int c, unsigned int n)
{
	__memset_64((uint64_t) dst, c, (unsigned int) n);
}

static int
is_this_valid_kseg0_addr(uint64_t _paddr, uint64_t _size)
{
	unsigned long paddr = (unsigned long) _paddr;
	unsigned long size = (unsigned long) _size;

	paddr = paddr & 0x1fffffff;
	if ((paddr >= xlr_lib_kseg0_start)
		&& ((paddr + size) < (xlr_lib_kseg0_start + xlr_lib_kseg0_size))) {
		return 1;
	}

	/* is it is in the nmi region */
	if (paddr >= 0x1f100000 && ((paddr + size) <= 0x1fffffff))
		return 1;

	return 0;
}

static uint64_t
elf_get_vaddr(struct elf_binary *elf, uint64_t paddr, int lib_run_mode)
{
	if (lib_run_mode == KSEG0_APP) {
		return (paddr & 0x1fffffffULL);
	} else {
		return paddr;
	}
}

static int
is_bootloader_region(uint64_t paddr, uint64_t size, int lib_run_mode)
{
	/*Foll. Condition may evalute to true only in case of KSEG0 applications */
	if (lib_run_mode == KSEG0_APP) {
		if (!is_this_valid_kseg0_addr(paddr, size))
			return 1;
	}
	return 0;
}

/*****************************************************************************************/
static int
xlr_lib_copy_image(struct elf_binary *elf, int lib_run_mode,
				   int loading_nmi_image)
{
	uint64_t paddr;
	int tlb_entry, i = 0, j = 0;
	int retval = 0, cpu_index, mapped_tlbs;
	Elf_ExtPhdr *segment = NULL;
	struct xlr_vcpu_wakeup_info *t = NULL;
	unsigned int tlb_size;
	unsigned long long allocated_size;
	unsigned long long vaddr;
	struct ext_tlb *tlb_ptr;
	struct cpu_tlb_mapping *p;

	if (lib_verbose)
		print_segment_map(elf, 0);

	if (loading_nmi_image) {
		for (i = 0; i < nseg; i++) {
			segment = &segment_map[i];
			if (!init_done(segment)) {
				retval =
					load_elf_segment(elf, segment, 0, memsz(elf, segment),
									 paddr(elf, segment), lib_run_mode,
									 loading_nmi_image);
				if (retval != 0) {
					printf("load_elf_segment_64() failure\n");
					return -1;
				}
				segment->init_done = 1;
			}
		}
		return 0;
	}

	if (lib_run_mode == KSEG0_APP) {
		t = (struct xlr_vcpu_wakeup_info *) (xlr_vcpu_wakeup_info +
											 kseg_master_cpu);
		memset(&t->seg_info[0], 0,
			   sizeof (t->seg_info[0]) * MAX_ELF_SEGMENTS);
		memset(&t->map[0], 0, sizeof (t->map[0]) * MAX_TLB_MAPPINGS);
		t->printk_lock = (lib_spinlock_t *) & print_lock;

		for (i = 0; i < nseg; i++) {
			/* Load all segments. Do not bother about tlb entries;
			 * This is a kseg0 app. Setup wakeup_info struct.
			 */
			segment = &segment_map[i];
			if (!init_done(segment)) {
				retval =
					load_elf_segment(elf, segment, 0, memsz(elf, segment),
									 paddr(elf, segment), lib_run_mode,
									 loading_nmi_image);
				if (retval != 0) {
					printf("load_elf_segment_64() failure\n");
					return -1;
				}
				segment->init_done = 1;
			}
			/*Useful For Coredump. In "userapp" foll. values will be copied in 
			   to coredump datastructure. */
			t->seg_info[i].vaddr = elf_uval(elf, &segment->hdr, p_vaddr);
			t->seg_info[i].memsz = elf_uval(elf, &segment->hdr, p_memsz);
			t->seg_info[i].flags = elf_uval(elf, &segment->hdr, p_flags);
		}

		if (ELFCLASS64 == elf->class)
			t->func = (unsigned long) elf_uval(elf, elf->ehdr, e_entry);
		else {
			if (kseg2_image)
				t->func = elf_uval(elf, elf->ehdr, e_entry) & 0xbfffffff;
			else
				t->func = elf_uval(elf, elf->ehdr, e_entry);
		}
		t->valid = 1;
		t->map[0].phys = xlr_lib_kseg0_start;
		t->map[0].page_size = xlr_lib_kseg0_size;
		return 0;
	}

	if (lib_run_mode != KUSEG_APP)
		return 0;

	for (cpu_index = 0; cpu_index < 32; cpu_index++) {
		if (!(kuseg_master_cpu_mask & (1U << cpu_index)))
			continue;

		t = (struct xlr_vcpu_wakeup_info *) (xlr_vcpu_wakeup_info + cpu_index);
		p = t->map;
		t->map_count = 0;
		tlb_size = 0;
		tlb_entry = 0;

		memset(&t->seg_info[0], 0,
			   sizeof (t->seg_info[0]) * MAX_ELF_SEGMENTS);
		memset(&t->map[0], 0, sizeof (t->map[0]) * MAX_TLB_MAPPINGS);

		for (i = 0; i < nseg; i++) {
			segment = &segment_map[i];
			if (!init_done(segment)) {
				/*Either we are enterd here first time or this segment 
				   is not shared between all master cpus. */
				segment->ntlb = 0;
				for (j = 0, allocated_size = 0;
					 allocated_size < segment_size(elf, segment); ++j) {
					vaddr = vaddr(elf, segment) + allocated_size;
					tlb_size =
						best_fit_tlb_size(vaddr,
										  segment_size(elf,
													   segment) -
										  allocated_size);
					if (contig_alloc_off != 0
						&& paddr(elf, segment) == contig_alloc_off) {
						paddr = paddr(elf, segment) + allocated_size;
					} else {
						paddr = find_memblk(2 * tlb_size);
						if (!paddr) {
							printf("ENOMEM: while loading on cpu %d\n",
								   cpu_index);
							free_tlb_list_nomem(success_mask ?
												FREE_TLB_NONSHARED :
												FREE_TLB_ALL);
							return -1;
						}
					}

					segment->map[j].virt = vaddr;
					segment->map[j].phys = paddr;
					segment->map[j].page_size = tlb_size;
					if (ELFCLASS64 == elf->class)
						segment->map[j].asid = vaddr >> 32;
					else {
						if (kseg2_image)
							segment->map[j].asid = 0xffffffff;
						else
							segment->map[j].asid = 0;
					}
					segment->map[j].coherency = 3;
					segment->map[j].attr =
						(1 << 2) | (1 << 1) | (1 << 0) |
						(is_shared(segment) << BIT_SHARED);

					curr_tlb_mapping.virt = vaddr;
					curr_tlb_mapping.phys = paddr;
					curr_tlb_mapping.page_size = tlb_size;
					if (ELFCLASS64 == elf->class)
						curr_tlb_mapping.asid = 0;
					else {
						if (kseg2_image)
							curr_tlb_mapping.asid = 0xffffffff;
						else
							curr_tlb_mapping.asid = 0;
					}
					curr_tlb_mapping.coherency = 3;
					curr_tlb_mapping.attr =
						(1 << 2) | (1 << 1) | (1 << 0) |
						(is_shared(segment) << BIT_SHARED);

					if (build_tlb_list(&curr_tlb_mapping, i))
						return -1;
					++(segment->ntlb);
					/*printf(" tlb virt %lx phys %lx page %lx attr %x \n",
					   vaddr, paddr, tlb_size, segment->map[j].attr); */

					if (contig_alloc_off != 0
						&& paddr(elf, segment) == contig_alloc_off) {
						allocated_size += 2 * tlb_size;
						continue;
					}

					retval =
						load_elf_segment(elf, segment, allocated_size,
										 2 * tlb_size, paddr, lib_run_mode,
										 loading_nmi_image);
					if (retval != 0) {
						free_tlb_list_nomem(success_mask ?
											FREE_TLB_NONSHARED :
											FREE_TLB_ALL);
						printf("load_elf_segment_64()failure\n");
						return -1;
					}
					allocated_size += 2 * tlb_size;
				}
				segment->init_done = is_shared(segment) ? 1 : 0;
			}
			/*Useful For Coredump. Foll. values will be copied in to 
			   coredump datastructure. */
			t->seg_info[i].vaddr = elf_uval(elf, &segment->hdr, p_vaddr);
			t->seg_info[i].memsz = elf_uval(elf, &segment->hdr, p_memsz);
			t->seg_info[i].flags = elf_uval(elf, &segment->hdr, p_flags);
		}
		/*Setting The ELF ENTRY */
		t->func = elf_uval(elf, elf->ehdr, e_entry);
		t->valid = 1;

		for (i = 0, tlb_entry = 0; i < nseg; i++) {
			segment = &segment_map[i];
			for (j = 0; j < segment->ntlb; ++j, ++tlb_entry) {
				(p + tlb_entry)->virt = segment->map[j].virt;
				(p + tlb_entry)->phys = segment->map[j].phys;
				(p + tlb_entry)->page_size = segment->map[j].page_size;
				(p + tlb_entry)->asid = segment->map[j].asid;
				(p + tlb_entry)->coherency = segment->map[j].coherency;
				(p + tlb_entry)->attr = segment->map[j].attr;
				t->map_count++;
				if (t->map_count >= MAX_TLB_MAPPINGS)
					break;
			}
		}
		if (lib_verbose) {
			print_segment_map(elf, cpu_index);
		}
		mapped_tlbs = get_mapped_tlb_num();
		tlb_ptr = combine_tlb_list();

		if ((retval = check_tlb_condition(tlb_ptr, mapped_tlbs))) {
			if (retval == -2) {
				printf("Not enough tlbs to launch rmios\n");
			} else {
				printf("Error: Cannot accommodate _endaddr in ");
				printf("%d wired entries on cpu %d\n", MAX_WIRED_ENTRIES,
					   cpu_index);
			}
			free_tlb_list_nomem(success_mask ? FREE_TLB_NONSHARED :
								FREE_TLB_ALL);
			free(tlb_ptr);
			return -1;
		}
		success_mask = success_mask | (1U << cpu_index);
		release_exclusive_tlb_list();
		fill_per_cpu_tlb_info(tlb_ptr, cpu_index, mapped_tlbs);
	}
	cleanup_tlb_list();
	return 0;
}

static int
load_elf_segment(struct elf_binary *elf, Elf_ExtPhdr * segment,
				 unsigned long segoff, size_t size, uint64_t paddr,
				 int lib_run_mode, int loading_nmi_image)
{
	uint64_t dst;
	size_t nbytes;
	uint64_t src;

	if ((segoff + size) > memsz(elf, segment)) {
		size = memsz(elf, segment) - segoff;
	}

	dst = elf_get_vaddr(elf, paddr, lib_run_mode);
	//printf("dst = 0x%llx\n", (unsigned long long) dst);

	if (is_bootloader_region(dst, size, lib_run_mode) && !loading_nmi_image) {
		printf("trespassed into bootloader region\n");
		return -1;
	}

	if ((segoff + size) <= filesz(elf, segment)) {
		nbytes = size;
	} else if (segoff >= filesz(elf, segment)) {
		nbytes = 0;
	} else {
		nbytes = filesz(elf, segment) - segoff;
	}

	if (nbytes) {
		src = (uint64_t) elf->image + offset(elf, segment) + segoff;
		copy_data(dst, src, nbytes);
	}

	dst += nbytes;
	nbytes = size - nbytes;

	if (nbytes)
		set_data(dst, 0, nbytes);

	return 0;
}


/*This function will allocate the physical memory, and will fill tlb 
  information (for kuseg only), virt addr information in per cpu data structure.*/

static int
check_segment(struct elf_binary *elf, const elf_phdr * phdr,
			  int lib_run_mode, int loading_nmi_image)
{
	uint64_t paddr = 0, memsz = 0;

	paddr = elf_uval(elf, phdr, p_paddr);
	memsz = elf_uval(elf, phdr, p_memsz);

	if ((!override_checks) && (!loading_nmi_image)) {
		if (lib_run_mode == KSEG0_APP) {
			if (!is_this_valid_kseg0_addr(paddr, memsz)) {
				if (((ELFCLASS64 == elf->class)
					 && paddr < 0xffffffff80000000ULL)
					|| paddr < 0x80000000) {
					printf("elf_download():loader error.\n");
					printf("Program been compiled for KUSEG");
					printf(" but being run in KSEG0 mode\n");
				} else {
					printf("elf_download():loader error.\n");
					printf("Invalid KSEG0 Address\n");
				}
				return -2;
			}
		} else
			if (((ELFCLASS64 == elf->class) && paddr >= 0x3fffffffffffffffULL)
			        || (paddr >= 0x80000000 && paddr < 0xc0000000)) {
                          printf("elf_download(): loader error. Program might have been"
				  "compiled for KSEG0 but being run in KUSEG mode\n");
			return -2;
		}
	}
	if (loading_nmi_image) {
		if (((ELFCLASS64 == elf->class) && (paddr < 0xffffffffbf100000ULL || paddr > 0xffffffffbfffffffULL)) 
                    || (paddr < 0xbf100000 || paddr > 0xbfffffff)) {
			printf ("\nelf_download(): loader error. Cant do nmiload on this image\n");
			return -2;
		}
	}
	return 0;
}

static int
add_runtime_segs(struct elf_binary *elf)
{
	int tlb_size;
	uint32_t p_align;
	unsigned long long vaddr;
	elf_phdr *hdr;
	extern int contig_memory_init(char *vaddr, unsigned long long paddr,
								  unsigned long long size);


	/* Add heap to the end of the last seg */
	hdr = &(segment_map[nseg - 1].hdr);

    /* Save the app endaddr */
    app_endaddr =  elf_uval(elf, hdr, p_vaddr) + elf_uval(elf, hdr, p_memsz); 

	elf_hdr_elm(elf, hdr, p_memsz,
				(elf_uval(elf, hdr, p_memsz) + heap_size));

	if (contig_alloc_off == 0 || contig_alloc_sz == 0)
		return 0;

	/* Add shared mem with paddr, find out the paddr alignment and 
	   create a vaddr for that alignment  */
	tlb_size = best_fit_tlb_size(contig_alloc_off, contig_alloc_sz);
	if (!tlb_size) {
		printf
			("Error, Could align to a tlb for shared mem offset and size\n");
		return -1;
	}
	p_align = 2 * tlb_size;

	/* Align the vaddr to p align */
	vaddr =
		nlm_roundup((elf_uval(elf, hdr, p_vaddr) +
					 elf_uval(elf, hdr, p_memsz)), p_align);

	hdr = &(segment_map[nseg].hdr);
	segment_map[nseg].init_done = 0;
	segment_map[nseg].is_shared = 1;
	elf_hdr_elm(elf, hdr, p_vaddr, vaddr);
	elf_hdr_elm(elf, hdr, p_paddr, contig_alloc_off);
	elf_hdr_elm(elf, hdr, p_memsz, contig_alloc_sz);
	elf_hdr_elm(elf, hdr, p_align, p_align);
	nseg++;

	contig_memory_init((char *) (long) vaddr, contig_alloc_off,
					   contig_alloc_sz);
	return 0;
}

static int
is_mergable(struct elf_binary *elf, const elf_phdr * segment,
			const elf_phdr * phdr)
{
	int ret;

	ret = (elf_uval(elf, phdr, p_type) == elf_uval(elf, segment, p_type));
	ret +=(elf_uval(elf, phdr, p_offset) ==
		 elf_uval(elf, segment, p_offset) + elf_uval(elf, segment, p_memsz));
	ret += (elf_uval(elf, phdr, p_filesz) == elf_uval(elf, segment, p_memsz));
	ret += (elf_uval(elf, phdr, p_vaddr) ==
		 elf_uval(elf, segment, p_vaddr) + elf_uval(elf, segment, p_memsz));
	ret += (elf_uval(elf, phdr, p_paddr) ==
		 elf_uval(elf, segment, p_paddr) + elf_uval(elf, segment, p_memsz));
	ret += (elf_uval(elf, phdr, p_flags) == elf_uval(elf, segment, p_flags));

	return (ret == 6);
}

static int
add_segment(struct elf_binary *elf, const elf_phdr * phdr,
			int lib_run_mode)
{
	Elf_ExtPhdr *segment;

	if (nseg >= MAX_ELF_SEGMENTS) {
		printf("Too many segment\n");
		return -1;
	}

	/* check if the segment can be merged with the previous segment. Currently,
	   only shared bss segment can be merged with the previous one, the shared
	   data segment. The shared heap segment doesn't appear as a segment in the
	   elf executable and is folded into the shared segment by adding its size
	   to the memory size of the shared segment
	 */

	if (nseg) {
		segment = &segment_map[nseg - 1];
		if (is_mergable(elf, &segment->hdr, phdr)) {
			elf_hdr_elm(elf, &segment->hdr, p_memsz,
						(memsz(elf, segment) +
						 elf_uval(elf, phdr, p_memsz)));
			//segment_map[nseg-1].hdr.p_memsz += phdr64->p_memsz;
			elf_hdr_elm(elf, &segment->hdr, p_filesz,
						(filesz(elf, segment) +
						 elf_uval(elf, phdr, p_filesz)));
			//    segment_map[nseg-1].hdr.p_filesz += phdr64->p_filesz;
			return 0;
		}
	}
	//segment_map[nseg].hdr = (*phdr);
	memcpy_virt_phys((unsigned char *) &segment_map[nseg].hdr,
					 (uint64_t) phdr, sizeof (elf_phdr));
	segment_map[nseg].init_done = 0;
	segment_map[nseg].is_shared = 1;

	if (lib_run_mode == KUSEG_APP) {
		int can_share = ((elf_uval(elf, phdr, p_flags) & PF_SHR)
						 || (is_code_segment(elf, phdr)
							 && commoncode_seg_enabled));
		if (!can_share) {
			segment_map[nseg].is_shared = 0;
		}
	}

	++nseg;
	return 0;
}

static int
patch_segments(struct elf_binary *elf, int lib_run_mode)
{
	int i;
	elf_phdr *phdr;
	unsigned int tlb_size;
	uint64_t paddr, align, memsz, flags;

	for (i = 0; i < nseg; ++i) {
		phdr = &segment_map[i].hdr;

		paddr = elf_uval(elf, phdr, p_paddr);
		memsz = elf_uval(elf, phdr, p_memsz);
		align = elf_uval(elf, phdr, p_align);
		flags = elf_uval(elf, phdr, p_flags);

		if (lib_run_mode == KUSEG_APP) {
			tlb_size = best_fit_tlb_size(paddr, paddr);
			if (!tlb_size) {
				// error
			}
			elf_hdr_elm(elf, phdr, p_align, (2 * tlb_size));
		}
#if 0
		if (flags & PF_SHR) {
			elf_hdr_elm(elf, phdr, p_memsz,
						memsz + exec_info_64.shared_heap_size_64);
			//phdr->p_memsz += exec_info_64.shared_heap_size_64;
		}
#endif
	}

	return 0;
}

int
elf_download(struct elf_binary *elf, const int lib_run_mode,
			 const int loading_nmi_image)
{
	int i;
	int retval = 0;
	uint64_t e_entry = 0, e_phnum = 0;
	const elf_phdr *phdr;

	nseg = 0;
	e_entry = elf_uval(elf, elf->ehdr, e_entry);
	e_phnum = elf_uval(elf, elf->ehdr, e_phnum);

	/* Do the address map checkings for 64 bit addressing here */
	if (lib_run_mode == KUSEG_APP) {
		if (e_entry > 0x3fffffffffffffffULL) {
			printf("Error: Program is compiled for KUSEG\n");
			printf("but falls out of xuseg region:");
			printf(" 0x0 - 0x3fffffffffffffff\n");
			return -1;
		}
	} else {
		if (e_entry < 0xffffffff80000000ULL) {
			printf("Error: Program is compiled for KSEG\n");
			printf(" but being run in KUSEG mode\n");
			return -1;
		}
	}

	for (i = 0; i < e_phnum; i++) {
		phdr = elf_phdr_by_index(elf, i);
		if (!elf_phdr_is_loadable(elf, phdr))
			continue;

		/* Instances when people compile an rmios application for kuseg
		 * mode and run in kseg0 mode and vice-versa are frequent. In
		 * such instances the following call returns with an error
		 */
		retval = check_segment(elf, phdr, lib_run_mode, loading_nmi_image);
		if (retval < 0) {
			return retval;
		}
		/* record segment information from elf file into extended 
		 * elf headers. Segments, if they can be, are merged with
		 * previous segments.
		 */

		retval = add_segment(elf, phdr, lib_run_mode);
		if (retval < 0) {
			printf("error: add_segment() returned %d\n", retval);
			return retval;
		}

	}
	retval = patch_segments(elf, lib_run_mode);
	if (retval) {
		return retval;
	}

	/* Add the runtime segs such as heap and shared mem */
	if ((retval = add_runtime_segs(elf)) != 0)
		return retval;

	/*@ here, we have segment_map_64 structure ready, it contains information 
	   about each and every segment present in elf file. Foll. Function will 
	   allocate required physical memory for all segements and will copy the 
	   image. */
	retval = xlr_lib_copy_image(elf, lib_run_mode, loading_nmi_image);
	if (retval)
		return -1;
	return 0;
}
