#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "mm.h"
#include "memblk.h"
#include "xlr_boot_lib.h"
#include "elf-ext.h"

/* Data structure tlb_info[segs] stores tlb information of segs into resp tlb_info
 * entry. Every time a tlb is allocated, it goes into one of the segments... The
 * segment can be shared amongnst cpus or it is exclusive to a cpu
 *
 * Following functions act on this data structure in different ways to
 * build/cleanup tlb lists.
 *
 * 1. int build_tlb_list(struct ext_tlb *tlb, int seg_num)
 *
 * 2. struct ext_tlb *combine_tlb_list()
 *
 * 3. When a cpu is stopped, memory is released back into mem pull 
 *    returns amount of memory reclaimed
 *    uint64_t free_per_cpu_tlb_list(int cpu_index, int flag)
 *   
 * 4. Reinitialize tlb list 
 *    void cleanup_tlb_list()
 *
 * 5. Failure case: In case no memory is found for segments to be loaded, this
 *    function releases allocated memory into mem pull
 *
 *    Add into memblock whatever tlbs are not shared or everything if this is the 
 *    only cpu.
 *    void free_tlb_list_nomem(int flag)
 *
 * 6. void release_exclusive_tlb_list()
 *
 * 7. int get_mapped_tlb_num()
 */

#define BIT_SHARED 3

/* Global Variables */
static struct tlb_info {
	int num_tlbs;
	int shared;
	struct ext_tlb *page_list;
} tlb_info[MAX_ELF_SEGMENTS];
static long long wired_endaddr;

/* Extern Variables */
extern struct per_cpu_info_loader per_cpu_info[MAX_CPUS];
extern Elf_ExtPhdr segment_map[MAX_ELF_SEGMENTS];

/* Extern Functions */
extern void *kseg_to_mmap(unsigned long long);
extern void coalesce_memblk();
extern unsigned long long mmap_to_kseg(void *);

/******************************************************************************************/
int
build_tlb_list(struct ext_tlb *tlb, int seg_num)
{
	struct ext_tlb *curr = NULL;
	struct ext_tlb *this_page = NULL;
	int this_num;

	this_page = tlb_info[seg_num].page_list;
	this_num = tlb_info[seg_num].num_tlbs;

#ifdef DEBUG
	printf("in build_tlb_list ... seg num is %d\n", seg_num);
	printf("tlb #%d: virt = %llx, phys = %llx, size = %x attr = %x\n",
		   this_num, tlb->virt, tlb->phys, tlb->page_size, tlb->attr);
#endif							/* DEBUG */

	/* If condition mallocs additional MAX_TLB_MAPPINGS entries if the
	 * already allocated buffer is either null or full... If the previous
	 * buffer hold some entries, those are copied to newly malloced struct
	 * realloc???
	 */

	if (!(this_num % MAX_TLB_MAPPINGS)) {
		curr = (struct ext_tlb *) malloc(sizeof (struct ext_tlb) *
										 (this_num + MAX_TLB_MAPPINGS));
		if (!curr) {
			printf("Insufficient memory to generate page list\n");
			return -1;
		}
		if (!this_page)
			this_page = curr;
		else {
			memcpy(curr, this_page, sizeof (struct ext_tlb)
				   * this_num);
			free(this_page);
			this_page = curr;
		}
	}

	if (!this_page) {
		printf("something wrong in build_tlb_list... fix it.\n");
		return -1;
	}

	curr = this_page + this_num;
	memcpy(curr, tlb, sizeof (struct ext_tlb));
	tlb_info[seg_num].page_list = this_page;
	tlb_info[seg_num].num_tlbs++;

	if (tlb->attr & (1 << BIT_SHARED))
		tlb_info[seg_num].shared = 1;
	else
		tlb_info[seg_num].shared = 0;

	return 0;

}

int
get_mapped_tlb_num()
{
	int total_tlbs = 0, i;

	for (i = 0; i < MAX_ELF_SEGMENTS; i++)
		total_tlbs += tlb_info[i].num_tlbs;
	return total_tlbs;
}

/* combines the tlbs in different segments, and reinitializes tlbs for
 * the segments which are exclusively mapped for this cpu
 */
struct ext_tlb *
combine_tlb_list()
{
	struct ext_tlb *tlb_ptr;
	int total_tlbs = 0, i;
#ifdef DEBUG
	struct ext_tlb *temp;
#endif							/* DEBUG */

	for (i = 0; i < MAX_ELF_SEGMENTS; i++)
		total_tlbs += tlb_info[i].num_tlbs;

	if (!total_tlbs) {
		printf("oops... number of tlbs is 0000.. in combine_tlb_list\n");
		return 0;
	}

	tlb_ptr =
		(struct ext_tlb *) malloc(total_tlbs * sizeof (struct ext_tlb));
	if (!tlb_ptr) {
		printf("Cant get persistent memory... fix it\n");
		return NULL;
	}
#ifdef DEBUG
	printf("in combine_tlb_list, num_tlb is %d\n", total_tlbs);
#endif							/* DEBUG */

	for (i = 0; i < MAX_ELF_SEGMENTS; i++) {

		memcpy(tlb_ptr, tlb_info[i].page_list, sizeof (struct ext_tlb) *
			   tlb_info[i].num_tlbs);
		tlb_ptr += tlb_info[i].num_tlbs;
	}

	tlb_ptr -= total_tlbs;

#ifdef DEBUG
	temp = tlb_ptr;
	for (i = 0; i < total_tlbs; i++) {
		printf("tlb #%d: virt = %llx, phys = %llx, size = %x attr = "
			   "%x\n", i, temp->virt, temp->phys, temp->page_size,
			   temp->attr);
		temp++;
	}
#endif							/* DEBUG */

	return tlb_ptr;
}

/* When a cpu is stopped, memory is released back into mem pull */
/* returns amount of memory reclaimed */
uint64_t
free_per_cpu_tlb_list(int cpu_index, int flag)
{
	struct ext_tlb *tlb_ptr = NULL;
	volatile struct per_cpu_info_loader *this_per_cpu_info = per_cpu_info;
	uint64_t mem = 0;
	int num_tlbs;
	int i;

	if (!((flag == FREE_TLB_ALL) || (flag == FREE_TLB_NONSHARED))) {
		printf("Incorrect flag in free_per_cpu_tlb_list\n");
		return 0;
	}

	this_per_cpu_info = per_cpu_info + cpu_index;

	num_tlbs = (int) this_per_cpu_info->num_tlbs;
	//    tlb_ptr = (struct ext_tlb *)(long)this_per_cpu_info->ext_tlb_ptr;
	tlb_ptr =
		(struct ext_tlb *) kseg_to_mmap(this_per_cpu_info->ext_tlb_ptr);
	if (!tlb_ptr) {
		printf("Error!!! free_per_cpu_tlb_list tlb should not be null\n");
		return 0;
	}
#ifdef DEBUG
	printf("freeing mem in for loop for cpu %x\n", cpu_index);
#endif							/* DEBUG */
	for (i = 0; i < num_tlbs; i++) {
#ifdef DEBUG
		printf("flag = %d, attr = %d\n", flag, tlb_ptr->attr);
		printf("phys = %llx, size = %x\n", tlb_ptr->phys,
			   2 * tlb_ptr->page_size);
#endif							/* DEBUG */
		if ((flag == FREE_TLB_NONSHARED)
			&& (tlb_ptr->attr & (1 << BIT_SHARED))) {
			tlb_ptr++;
			continue;
		}
#ifdef DEBUG
		printf("Freeing mem at phys = %llx, size = %x\n", tlb_ptr->phys,
			   2 * tlb_ptr->page_size);
#endif							/* DEBUG */
		add_memblk(tlb_ptr->phys, 2 * tlb_ptr->page_size);
		mem += 2 * tlb_ptr->page_size;
		tlb_ptr++;
	}
	//    tlb_ptr = (struct ext_tlb *)(long)this_per_cpu_info->ext_tlb_ptr;
	tlb_ptr =
		(struct ext_tlb *) kseg_to_mmap(this_per_cpu_info->ext_tlb_ptr);
	free((void *) tlb_ptr);
	this_per_cpu_info->num_tlbs = 0;
	this_per_cpu_info->ext_tlb_ptr = 0;
	return mem;
}

/* Reinitialize tlb list */
void
cleanup_tlb_list()
{
	int i;

	for (i = 0; i < MAX_ELF_SEGMENTS; i++) {
		if (tlb_info[i].page_list) {
			free(tlb_info[i].page_list);
			tlb_info[i].page_list = NULL;
		}
		tlb_info[i].shared = 0;
		tlb_info[i].num_tlbs = 0;
	}
}

/* Release tlb list of the segments exclusive to a cpu */
void
release_exclusive_tlb_list()
{
	int i;

	for (i = 0; i < MAX_ELF_SEGMENTS; i++) {
		if (!tlb_info[i].page_list)
			continue;
		/* Put memory allocated to non-shared tlbs back to mem pull */

		if (!tlb_info[i].shared) {
			free(tlb_info[i].page_list);
			tlb_info[i].page_list = NULL;
			tlb_info[i].shared = 0;
			tlb_info[i].num_tlbs = 0;
		}
	}
}

/* Failure case: In case no memory is found for segments to be loaded, this
 * function releases allocated memory into mem pull
 */
/* Also, if the _endaddr doesnt fall within wired_entries, this is called */
/* Add into memblock whatever tlbs are not shared or everything if this is the 
 * only cpu.
 */
void
free_tlb_list_nomem(int flag)
{
	int i, j;
	struct ext_tlb *curr;

	if (!((flag == FREE_TLB_ALL) || (flag == FREE_TLB_NONSHARED))) {
		printf("Incorrect flag in free_tlb_list_nomem \n");
		return;
	}

	for (i = 0; i < MAX_ELF_SEGMENTS; i++) {
		if (!tlb_info[i].page_list)
			continue;

		/* Put memory allocated to non-shared tlbs back to mem pull */

		if ((flag == FREE_TLB_ALL) || ((!tlb_info[i].shared) && (flag ==
																 FREE_TLB_NONSHARED)))
		{
			curr = tlb_info[i].page_list;
			for (j = 0; j < tlb_info[i].num_tlbs; j++) {
				add_memblk(curr->phys, 2 * curr->page_size);
				curr++;
			}
		}
	}

	/* Reset ntlb of all segments to 0 */
	for (i = 0; i < MAX_ELF_SEGMENTS; i++) {
		segment_map[i].ntlb = 0;
	}

	coalesce_memblk();
	/* Finally clean up all tlb lists */
	cleanup_tlb_list();
}

/* checks if _endaddr falls within MAX_WIRED_ENTRIES tlbs... 
 * If not, report failure. We cannot launch this image.
 * 0 -> success
 * -1 -> failure - wired endaddr condition failed.
 * -2 -> failure - tlb refill handler not supported in rmios app
 */
int
check_tlb_condition(struct ext_tlb *tlb_ptr, int num_tlbs)
{
	int i;

	/* success... return 0 */
	if (num_tlbs <= MAX_TLB_MAPPINGS)
		return 0;
#if 0
	/* check version of this rmios app
	   pre 1.6 do not have tlb refill handler.
	   return error if pre 1.6 rmios app is found
	 */
	if (version_info.version < VERSION_1_6)
		return -2;
#endif
	/* Could not get any tlb mapping */
	if (!tlb_ptr)
		return -1;

	for (i = MAX_WIRED_ENTRIES + 1, tlb_ptr += MAX_WIRED_ENTRIES; i <= num_tlbs; i++) {
		if (wired_endaddr >= tlb_ptr->virt)
			return -1;
		tlb_ptr++;
	}
	/* success... return 0 */
	return 0;
}

/* Copy the combined tlb list ptr to tlb entry pointer of
 * per_cpu_info struct 
 */
void
fill_per_cpu_tlb_info(struct ext_tlb *tlb_ptr, int cpu_index, int num_tlbs)
{
#ifdef DEBUG
	int i;
	struct ext_tlb *temp;
#endif							/* DEBUG */
	unsigned long long kseg_tlb_ptr;
	volatile struct per_cpu_info_loader *this_per_cpu_info = per_cpu_info;
	this_per_cpu_info = per_cpu_info + cpu_index;

	kseg_tlb_ptr = mmap_to_kseg(tlb_ptr);

	if (tlb_ptr) {
		this_per_cpu_info->num_tlbs = num_tlbs;
		this_per_cpu_info->ext_tlb_ptr = kseg_tlb_ptr;
	} else {
		this_per_cpu_info->num_tlbs = 0;
		this_per_cpu_info->ext_tlb_ptr = 0;
	}
#ifdef DEBUG
	temp = (struct ext_tlb *) kseg_to_mmap(this_per_cpu_info->ext_tlb_ptr);
	for (i = 0; i < num_tlbs; i++) {
		printf("tlb #%d: virt = %llx, phys = %llx, size = %x attr = %x\n",
			   i, temp->virt, temp->phys, temp->page_size, temp->attr);
		temp++;
	}
	temp = (struct ext_tlb *) kseg_to_mmap(this_per_cpu_info->ext_tlb_ptr);
	printf("kseg_to_mmap is %p", temp);
	printf(" this_per_cpu_info->ext_tlb_ptr is %llx\n",
		   this_per_cpu_info->ext_tlb_ptr);
	printf("this_per_cpu_info->ext_tlb_ptr long is %lx\n",
		   (long) this_per_cpu_info->ext_tlb_ptr);

#endif							/* DEBUG */
}

/* Copy tlb info of master cpu into buddy cpu */
void
copy_per_cpu_tlb_info(int master_cpu, int buddy_cpu)
{
	volatile struct per_cpu_info_loader *m, *b;

	m = per_cpu_info + master_cpu;
	b = per_cpu_info + buddy_cpu;

	b->num_tlbs = m->num_tlbs;
	b->ext_tlb_ptr = m->ext_tlb_ptr;
}


#ifdef NETOS_LOADER_TODO
static struct ext_tlb *
get_parent_ext_tlb(struct xlr_vcpu_wakeup_info *parent, uint64_t vaddr)
{
	struct ext_tlb *tlb_ptr = NULL;
	volatile struct per_cpu_info_loader *this_per_cpu_info = per_cpu_info;
	int num_tlbs;
	int i;

	this_per_cpu_info = per_cpu_info + parent->master_cpu;

	num_tlbs = (int) this_per_cpu_info->num_tlbs;

	tlb_ptr =
		(struct ext_tlb *) kseg_to_mmap(this_per_cpu_info->ext_tlb_ptr);

	if (!tlb_ptr) {
		printf("Error!!! free_per_cpu_tlb_list tlb should not be null\n");
		return 0;
	}

	for (i = 0; i < num_tlbs; i++) {
		if ((vaddr == tlb_ptr->virt)) {
			return tlb_ptr;
		}
		tlb_ptr++;
	}
	printf("Found no tlb_ptr !!!!");
	return NULL;
}
#endif

